# run states
RUN_IDLE = "idle"
RUN_INITIALIZED = "initialized"
RUN_USER_ACTION_NEEDED = "userActionNeeded"
RUN_USER_ACTION_COMPLETED = "ready"
RUN_STARTED = "started"
RUN_FAILED = "error"
RUN_COMPLETED = "done"
RUN_ABORTED = "aborted"
RUN_AUTHENTICATED = "authenticated"
RUN_DAEMON = "daemon"
